import copy
import inspect
import warnings

import dill
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

from agilerl.networks.evolvable_cnn import EvolvableCNN
from agilerl.networks.evolvable_mlp import EvolvableMLP
from agilerl.utils.algo_utils import chkpt_attribute_to_device, unwrap_optimizer
from agilerl.wrappers.make_evolvable import MakeEvolvable


class TD3:
    """The TD3 algorithm class. TD3 paper: https://arxiv.org/abs/1802.09477

    :param state_dim: State observation dimension
    :type state_dim: list[int]
    :param action_dim: Action dimension
    :type action_dim: int
    :param one_hot: One-hot encoding, used with discrete observation spaces
    :type one_hot: bool
    :param max_action: Upper bound of the action space, defaults to 1
    :type max_action: float, optional
    :param min_action: Lower bound of the action space, defaults to -1
    :type min_action: float, optional
    :param O_U_noise: Use Ornstein Uhlenbeck action noise for exploration. If False, uses Gaussian noise. Defaults to True
    :type O_U_noise: bool, optional
    :param vect_noise_dim: Vectorization dimension of environment for action noise, defaults to 1
    :type vect_noise_dim: int, optional
    :param expl_noise: Scale for Ornstein Uhlenbeck action noise, or standard deviation for Gaussian exploration noise
    :type expl_noise: float, optional
    :param mean_noise: Mean of exploration noise, defaults to 0.0
    :type mean_noise: float, optional
    :param theta: Rate of mean reversion in Ornstein Uhlenbeck action noise, defaults to 0.15
    :type theta: float, optional
    :param dt: Timestep for Ornstein Uhlenbeck action noise update, defaults to 1e-2
    :type dt: float, optional
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param net_config: Network configuration, defaults to mlp with hidden size [64,64]
    :type net_config: dict, optional
    :param batch_size: Size of batched sample from replay buffer for learning, defaults to 64
    :type batch_size: int, optional
    :param lr_actor: Learning rate for actor optimizer, defaults to 1e-4
    :type lr_actor: float, optional
    :param lr_critic: Learning rate for critic optimizer, defaults to 1e-3
    :type lr_critic: float, optional
    :param learn_step: Learning frequency, defaults to 5
    :type learn_step: int, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param tau: For soft update of target network parameters, defaults to 1e-3
    :type tau: float, optional
    :param mut: Most recent mutation to agent, defaults to None
    :type mut: str, optional
    :param policy_freq: Frequency of critic network updates compared to policy network, defaults to 2
    :type policy_freq: int, optional
    :param actor_network: Custom actor network, defaults to None
    :type actor_network: nn.Module, optional
    :param critic_networks: List of two custom critic networks (one for each of TD3's two critics), defaults to None
    :type critic_networks: list[nn.Module], optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    """

    def __init__(
        self,
        state_dim,
        action_dim,
        one_hot,
        max_action=1,
        min_action=-1,
        O_U_noise=True,
        vect_noise_dim=1,
        expl_noise=0.1,
        mean_noise=0.0,
        theta=0.15,
        dt=1e-2,
        index=0,
        net_config={"arch": "mlp", "hidden_size": [64, 64]},
        batch_size=64,
        lr_actor=1e-4,
        lr_critic=1e-3,
        learn_step=5,
        gamma=0.99,
        tau=0.005,
        mut=None,
        policy_freq=2,
        actor_network=None,
        critic_networks=None,
        device="cpu",
        accelerator=None,
        wrap=True,
    ):
        assert isinstance(
            state_dim, (list, tuple)
        ), "State dimension must be a list or tuple."
        assert isinstance(
            action_dim, (int, np.integer)
        ), "Action dimension must be an integer."
        assert isinstance(
            one_hot, bool
        ), "One-hot encoding flag must be boolean value True or False."
        assert isinstance(
            max_action,
            (float, int, np.float32, np.float64, np.integer, list, np.ndarray),
        ), "Max action must be a float or integer."
        assert isinstance(
            min_action,
            (float, int, np.float32, np.float64, np.integer, list, np.ndarray),
        ), "Min action must be a float or integer."
        if isinstance(min_action, list):
            assert (
                len(min_action) == action_dim
            ), "Length of min_action must be equal to action_dim."
            min_action = np.array(min_action)
        if isinstance(max_action, list):
            assert (
                len(max_action) == action_dim
            ), "Length of max_action must be equal to action_dim."
            max_action = np.array(max_action)
        if isinstance(max_action, np.ndarray) or isinstance(min_action, np.ndarray):
            assert np.all(
                max_action > min_action
            ), "Max action must be greater than min action."
        else:
            assert (
                max_action > min_action
            ), "Max action must be greater than min action."
        assert (isinstance(expl_noise, (float, int))) or (
            isinstance(expl_noise, np.ndarray)
            and expl_noise.shape == (vect_noise_dim, action_dim)
        ), "Exploration action noise rate must be a float, or an array of size action_dim"
        if isinstance(expl_noise, (float, int)):
            assert (
                expl_noise >= 0
            ), "Exploration noise must be greater than or equal to zero."
        assert isinstance(index, int), "Agent index must be an integer."
        assert isinstance(batch_size, int), "Batch size must be an integer."
        assert batch_size >= 1, "Batch size must be greater than or equal to one."
        assert isinstance(lr_actor, float), "Actor learning rate must be a float."
        assert lr_actor > 0, "Actor learning rate must be greater than zero."
        assert isinstance(lr_critic, float), "Critic learning rate must be a float."
        assert lr_critic > 0, "Critic learning rate must be greater than zero."
        assert isinstance(learn_step, int), "Learn step rate must be an integer."
        assert learn_step >= 1, "Learn step must be greater than or equal to one."
        assert isinstance(gamma, (float, int, torch.Tensor)), "Gamma must be a float."
        assert isinstance(tau, float), "Tau must be a float."
        assert tau > 0, "Tau must be greater than zero."
        assert isinstance(policy_freq, int), "Policy frequency must be an integer."
        assert (
            policy_freq >= 1
        ), "Policy frequency must be greater than or equal to one."
        if (actor_network is not None) != (
            critic_networks is not None
        ):  # XOR operation
            warnings.warn(
                "Actor and critic networks must both be supplied to use custom networks. Defaulting to net config."
            )
        assert isinstance(
            wrap, bool
        ), "Wrap models flag must be boolean value True or False."

        self.algo = "TD3"
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.one_hot = one_hot
        self.max_action = max_action
        self.min_action = min_action
        self.net_config = net_config
        self.batch_size = batch_size
        self.lr_actor = lr_actor
        self.lr_critic = lr_critic
        self.learn_step = learn_step
        self.gamma = gamma
        self.tau = tau
        self.mut = mut
        self.policy_freq = policy_freq
        self.O_U_noise = O_U_noise
        self.vect_noise_dim = vect_noise_dim
        self.expl_noise = (
            expl_noise
            if isinstance(expl_noise, np.ndarray)
            else expl_noise * np.ones((vect_noise_dim, action_dim))
        )
        self.mean_noise = (
            mean_noise
            if isinstance(mean_noise, np.ndarray)
            else mean_noise * np.ones((vect_noise_dim, action_dim))
        )
        self.current_noise = np.zeros((vect_noise_dim, action_dim))
        self.theta = theta
        self.dt = dt
        self.actor_network = actor_network
        self.critic_networks = critic_networks
        self.device = device
        self.accelerator = accelerator

        self.index = index
        self.scores = []
        self.fitness = []
        self.steps = [0]
        self.learn_counter = 0

        if self.actor_network is not None and self.critic_networks is not None:
            assert isinstance(
                critic_networks, (list, tuple)
            ), "Critic network must be a list or tuple"
            assert len(critic_networks) == 2, "TD3 requires exactly 2 critic networks."
            assert isinstance(critic_networks[0], type(actor_network)) and isinstance(
                critic_networks[1], type(actor_network)
            ), "'actor_network' and 'critic_networks' must be the same type."
            self.actor = actor_network
            self.critic_1, self.critic_2 = critic_networks
            if (
                isinstance(self.actor, (EvolvableMLP, EvolvableCNN))
                and isinstance(self.critic_1, (EvolvableMLP, EvolvableCNN))
                and isinstance(self.critic_1, (EvolvableMLP, EvolvableCNN))
            ):
                self.net_config = self.actor.net_config
            elif (
                isinstance(self.actor, MakeEvolvable)
                and isinstance(self.critic_1, MakeEvolvable)
                and isinstance(self.critic_1, MakeEvolvable)
            ):
                self.net_config = None
            else:
                assert (
                    False
                ), f"'actor_network' argument is of type {type(actor_network)} and critic networks are of type {type(critic_networks[0])}, \
                                both must be the same type and be of type EvolvableMLP, EvolvableCNN or MakeEvolvable"

        else:
            # model
            assert isinstance(self.net_config, dict), "Net config must be a dictionary."
            assert (
                "arch" in self.net_config.keys()
            ), "Net config must contain arch: 'mlp' or 'cnn'."

            if "mlp_output_activation" not in self.net_config.keys():
                if np.any(self.min_action < 0):
                    self.net_config["mlp_output_activation"] = "Tanh"
                else:
                    self.net_config["mlp_output_activation"] = "Sigmoid"

            critic_net_config = copy.deepcopy(self.net_config)
            critic_net_config["mlp_output_activation"] = (
                None  # Critic must have no output activation
            )

            if self.net_config["arch"] == "mlp":  # Multi-layer Perceptron
                assert (
                    "hidden_size" in self.net_config.keys()
                ), "Net config must contain hidden_size: int."
                assert isinstance(
                    self.net_config["hidden_size"], list
                ), "Net config hidden_size must be a list."
                assert (
                    len(self.net_config["hidden_size"]) > 0
                ), "Net config hidden_size must contain at least one element."
                # TD3 employs two critic networks
                self.actor = EvolvableMLP(
                    num_inputs=state_dim[0],
                    num_outputs=action_dim,
                    device=self.device,
                    accelerator=self.accelerator,
                    **self.net_config,
                )

                self.critic_1 = EvolvableMLP(
                    num_inputs=state_dim[0] + action_dim,
                    num_outputs=1,
                    device=self.device,
                    accelerator=self.accelerator,
                    **critic_net_config,
                )
                self.critic_2 = EvolvableMLP(
                    num_inputs=state_dim[0] + action_dim,
                    num_outputs=1,
                    device=self.device,
                    accelerator=self.accelerator,
                    **critic_net_config,
                )
            elif self.net_config["arch"] == "cnn":  # Convolutional Neural Network
                for key in [
                    "channel_size",
                    "kernel_size",
                    "stride_size",
                    "hidden_size",
                ]:
                    assert (
                        key in self.net_config.keys()
                    ), f"Net config must contain {key}: int."
                    assert isinstance(
                        self.net_config[key], list
                    ), f"Net config {key} must be a list."
                    assert (
                        len(self.net_config[key]) > 0
                    ), f"Net config {key} must contain at least one element."
                assert (
                    "normalize" in self.net_config.keys()
                ), "Net config must contain normalize: True or False."
                assert isinstance(
                    self.net_config["normalize"], bool
                ), "Net config normalize must be boolean value True or False."
                self.actor = EvolvableCNN(
                    input_shape=state_dim,
                    num_actions=action_dim,
                    device=self.device,
                    accelerator=self.accelerator,
                    **self.net_config,
                )
                self.critic_1 = EvolvableCNN(
                    input_shape=state_dim,
                    num_actions=action_dim,
                    critic=True,
                    device=self.device,
                    accelerator=self.accelerator,
                    **critic_net_config,
                )
                self.critic_2 = EvolvableCNN(
                    input_shape=state_dim,
                    num_actions=action_dim,
                    critic=True,
                    device=self.device,
                    accelerator=self.accelerator,
                    **critic_net_config,
                )

        self.actor_target = copy.deepcopy(self.actor)
        self.critic_target_1 = copy.deepcopy(self.critic_1)
        self.critic_target_2 = copy.deepcopy(self.critic_2)
        self.actor_target.load_state_dict(self.actor.state_dict())
        self.critic_target_1.load_state_dict(self.critic_1.state_dict())
        self.critic_target_2.load_state_dict(self.critic_2.state_dict())
        self.actor_optimizer = optim.Adam(self.actor.parameters(), lr=self.lr_actor)
        self.critic_1_optimizer = optim.Adam(
            self.critic_1.parameters(), lr=self.lr_critic
        )
        self.critic_2_optimizer = optim.Adam(
            self.critic_2.parameters(), lr=self.lr_critic
        )

        self.arch = (
            self.net_config["arch"] if self.net_config is not None else self.actor.arch
        )

        if self.accelerator is not None:
            if wrap:
                self.wrap_models()
        else:
            self.actor = self.actor.to(self.device)
            self.actor_target = self.actor_target.to(self.device)
            self.critic_1 = self.critic_1.to(self.device)
            self.critic_target_1 = self.critic_target_1.to(self.device)
            self.critic_2 = self.critic_2.to(self.device)
            self.critic_target_2 = self.critic_target_2.to(self.device)

        self.criterion = nn.MSELoss()

    def scale_to_action_space(self, action, convert_to_torch=False):
        """Scales actions to action space defined by self.min_action and self.max_action.

        :param action: Action to be scaled
        :type action: numpy.ndarray
        :param convert_to_torch: Flag to convert array to torch, defaults to False
        :type convert_to_torch: bool, optional
        """
        if convert_to_torch:
            max_action = (
                torch.from_numpy(self.max_action).to(self.device)
                if isinstance(self.max_action, (np.ndarray))
                else self.max_action
            )
            min_action = (
                torch.from_numpy(self.min_action).to(self.device)
                if isinstance(self.min_action, (np.ndarray))
                else self.min_action
            )
        else:
            max_action = self.max_action
            min_action = self.min_action

        mlp_output_activation = self.actor.mlp_output_activation
        if mlp_output_activation in ["Tanh"]:
            pre_scaled_min = -1
            pre_scaled_max = 1
        elif mlp_output_activation in ["Sigmoid", "Softmax"]:
            pre_scaled_min = 0
            pre_scaled_max = 1
        else:
            if (
                isinstance(max_action, (np.ndarray, torch.Tensor))
                and (np.inf not in max_action and -np.inf not in min_action)
            ) or (
                not isinstance(max_action, (np.ndarray, torch.Tensor))
                and (np.inf != max_action and -np.inf != min_action)
            ):
                action = (
                    torch.where(action > 0, action * max_action, action * -min_action)
                    if convert_to_torch
                    else np.where(action > 0, action * max_action, action * -min_action)
                )
            return action.clip(min_action, max_action)

        if not (
            isinstance(min_action, (np.ndarray, torch.Tensor))
            or isinstance(max_action, (np.ndarray, torch.Tensor))
        ):
            if pre_scaled_min == min_action and pre_scaled_max == max_action:
                return action

        return min_action + (max_action - min_action) * (action - pre_scaled_min) / (
            pre_scaled_max - pre_scaled_min
        )

    def multi_dim_clamp(self, min, max, input):
        """Multi-dimensional clamp function"""
        if not isinstance(min, np.ndarray) and not isinstance(max, np.ndarray):
            return torch.clamp(input, min, max)

        min = (
            torch.from_numpy(min).to(self.device)
            if isinstance(min, np.ndarray)
            else min
        )
        max = (
            torch.from_numpy(max).to(self.device)
            if isinstance(max, np.ndarray)
            else max
        )

        if isinstance(max, torch.Tensor) and isinstance(min, (int, float)):
            min = torch.full_like(max, min).to(self.device)
        if isinstance(min, torch.Tensor) and isinstance(max, (int, float)):
            max = torch.full_like(min, max).to(self.device)

        return torch.max(torch.min(input, max), min)

    def get_action(self, state, training=True):
        """Returns the next action to take in the environment.
        Epsilon is the probability of taking a random action, used for exploration.
        For epsilon-greedy behaviour, set epsilon to 0.

        :param state: Environment observation, or multiple observations in a batch
        :type state: numpy.ndarray[float]
        :param training: Agent is training, use exploration noise, defaults to True
        :type training: bool, optional
        """
        state = torch.from_numpy(state).float()
        if self.accelerator is None:
            state = state.to(self.device)
        else:
            state = state.to(self.accelerator.device)

        if self.one_hot:
            state = (
                nn.functional.one_hot(state.long(), num_classes=self.state_dim[0])
                .float()
                .squeeze()
            )

        if (self.arch == "mlp" and len(state.size()) < 2) or (
            self.arch == "cnn" and len(state.size()) < 4
        ):
            state = state.unsqueeze(0)

        self.actor.eval()
        with torch.no_grad():
            action_values = self.actor(state)
        self.actor.train()

        action = self.scale_to_action_space(action_values.cpu().data.numpy())

        if training:
            action = (action + self.action_noise()).clip(
                self.min_action, self.max_action
            )

        return action

    def action_noise(self):
        """Create action noise for exploration, either Ornstein Uhlenbeck or
            from a normal distribution.

        :return: Action noise
        :rtype: np.ndArray
        """
        if self.O_U_noise:
            noise = (
                self.current_noise
                + self.theta * (self.mean_noise - self.current_noise) * self.dt
                + self.expl_noise
                * np.sqrt(self.dt)
                * np.random.normal(size=(self.vect_noise_dim, self.action_dim))
            )
            self.current_noise = noise
        else:
            noise = np.random.normal(
                self.mean_noise,
                self.expl_noise,
                size=(self.vect_noise_dim, self.action_dim),
            )
        return noise.astype(np.float32)

    def reset_action_noise(self, indices):
        """Reset action noise."""
        self.current_noise[indices] = self.mean_noise[indices]

    def learn(self, experiences, noise_clip=0.5, policy_noise=0.2):
        """Updates agent network parameters to learn from experiences.

        :param experience: List of batched states, actions, rewards, next_states, dones in that order.
        :type experience: list[torch.Tensor[float]]
        :param noise_clip: Maximum noise limit to apply to actions, defaults to 0.5
        :type noise_clip: float, optional
        :param policy_noise: Standard deviation of noise applied to policy, defaults to 0.2
        :type policy_noise: float, optional
        """
        states, actions, rewards, next_states, dones = experiences
        if self.accelerator is not None:
            states = states.to(self.accelerator.device)
            actions = actions.to(self.accelerator.device)
            rewards = rewards.to(self.accelerator.device)
            next_states = next_states.to(self.accelerator.device)
            dones = dones.to(self.accelerator.device)

        if self.one_hot:
            states = (
                nn.functional.one_hot(states.long(), num_classes=self.state_dim[0])
                .float()
                .squeeze()
            )
            next_states = (
                nn.functional.one_hot(next_states.long(), num_classes=self.state_dim[0])
                .float()
                .squeeze()
            )

        if self.arch == "mlp":
            input_combined = torch.cat([states, actions], 1)
            q_value_1 = self.critic_1(input_combined)
            q_value_2 = self.critic_2(input_combined)
        elif self.arch == "cnn":
            q_value_1 = self.critic_1(states, actions)
            q_value_2 = self.critic_2(states, actions)

        with torch.no_grad():
            next_actions = self.actor_target(next_states)
            # Scale actions
            next_actions = self.scale_to_action_space(
                next_actions, convert_to_torch=True
            )
            noise = actions.data.normal_(0, policy_noise)
            if self.accelerator is not None:
                noise = noise.to(self.accelerator.device)
            else:
                noise = noise.to(self.device)
            noise = self.multi_dim_clamp(-noise_clip, noise_clip, noise)
            next_actions = next_actions + noise
            next_actions = self.multi_dim_clamp(
                self.min_action, self.max_action, next_actions
            )

            # Compute the target, y_j, making use of twin critic networks
            if self.arch == "mlp":
                next_input_combined = torch.cat([next_states, next_actions], 1)
                q_value_next_state_1 = self.critic_target_1(next_input_combined)
                q_value_next_state_2 = self.critic_target_2(next_input_combined)
            elif self.arch == "cnn":
                q_value_next_state_1 = self.critic_target_1(next_states, next_actions)
                q_value_next_state_2 = self.critic_target_2(next_states, next_actions)
            q_value_next_state = torch.min(q_value_next_state_1, q_value_next_state_2)

        y_j = rewards + ((1 - dones) * self.gamma * q_value_next_state)

        # Loss equation needs to be updated to account for two q_values from two critics
        critic_loss = self.criterion(q_value_1, y_j) + self.criterion(q_value_2, y_j)

        # critic loss backprop
        self.critic_1_optimizer.zero_grad()
        self.critic_2_optimizer.zero_grad()
        if self.accelerator is not None:
            self.accelerator.backward(critic_loss)
        else:
            critic_loss.backward()
        self.critic_1_optimizer.step()
        self.critic_2_optimizer.step()

        # update actor and targets every policy_freq learn steps
        self.learn_counter += 1
        if self.learn_counter % self.policy_freq == 0:
            policy_actions = self.actor.forward(states)
            policy_actions = self.scale_to_action_space(
                policy_actions, convert_to_torch=True
            )

            if self.arch == "mlp":
                input_combined = torch.cat([states, policy_actions], 1)
                actor_loss = -self.critic_1(input_combined).mean()
            elif self.arch == "cnn":
                actor_loss = -self.critic_1(states, policy_actions).mean()

            # actor loss backprop
            self.actor_optimizer.zero_grad()
            if self.accelerator is not None:
                self.accelerator.backward(actor_loss)
            else:
                actor_loss.backward()
            self.actor_optimizer.step()

            # Add in a soft update for both critic_targets
            self.soft_update(self.actor, self.actor_target)
            self.soft_update(self.critic_1, self.critic_target_1)
            self.soft_update(self.critic_2, self.critic_target_2)

            return actor_loss.item(), critic_loss.item()
        else:
            return None, critic_loss.item()

    def soft_update(self, net, target):
        """Soft updates target network."""
        for eval_param, target_param in zip(net.parameters(), target.parameters()):
            target_param.data.copy_(
                self.tau * eval_param.data + (1.0 - self.tau) * target_param.data
            )

    def test(self, env, swap_channels=False, max_steps=None, loop=3):
        """Returns mean test score of agent in environment with epsilon-greedy policy.

        :param env: The environment to be tested in
        :type env: Gym-style environment
        :param swap_channels: Swap image channels dimension from last to first [H, W, C] -> [C, H, W], defaults to False
        :type swap_channels: bool, optional
        :param max_steps: Maximum number of testing steps, defaults to None
        :type max_steps: int, optional
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean. Defaults to 3
        :type loop: int, optional
        """
        with torch.no_grad():
            rewards = []
            num_envs = env.num_envs if hasattr(env, "num_envs") else 1
            for i in range(loop):
                state, _ = env.reset()
                scores = np.zeros(num_envs)
                completed_episode_scores = np.zeros(num_envs)
                finished = np.zeros(num_envs)
                step = 0
                while not np.all(finished):
                    if swap_channels:
                        state = np.moveaxis(state, [-1], [-3])
                    action = self.get_action(state, training=False)
                    state, reward, done, trunc, _ = env.step(action)
                    step += 1
                    scores += np.array(reward)
                    for idx, (d, t) in enumerate(zip(done, trunc)):
                        if (
                            d or t or (max_steps is not None and step == max_steps)
                        ) and not finished[idx]:
                            completed_episode_scores[idx] = scores[idx]
                            finished[idx] = 1
                rewards.append(np.mean(completed_episode_scores))
        mean_fit = np.mean(rewards)
        self.fitness.append(mean_fit)
        return mean_fit

    def clone(self, index=None, wrap=True):
        """Returns cloned agent identical to self.

        :param index: Index to keep track of agent for tournament selection and mutation, defaults to None
        :type index: int, optional
        """
        input_args = self.inspect_attributes(input_args_only=True)
        input_args["wrap"] = wrap
        clone = type(self)(**input_args)

        if self.accelerator is not None:
            self.unwrap_models()
        actor = self.actor.clone()
        actor_target = self.actor_target.clone()
        critic_1 = self.critic_1.clone()
        critic_target_1 = self.critic_target_1.clone()
        critic_2 = self.critic_2.clone()
        critic_target_2 = self.critic_target_2.clone()

        actor_optimizer = optim.Adam(actor.parameters(), lr=clone.lr_actor)
        critic_1_optimizer = optim.Adam(critic_1.parameters(), lr=clone.lr_critic)
        critic_2_optimizer = optim.Adam(critic_2.parameters(), lr=clone.lr_critic)

        actor_optimizer.load_state_dict(self.actor_optimizer.state_dict())
        critic_1_optimizer.load_state_dict(self.critic_1_optimizer.state_dict())
        critic_2_optimizer.load_state_dict(self.critic_2_optimizer.state_dict())

        if self.accelerator is not None:
            if wrap:
                (
                    clone.actor,
                    clone.actor_target,
                    clone.critic_1,
                    clone.critic_target_1,
                    clone.critic_2,
                    clone.critic_target_2,
                    clone.actor_optimizer,
                    clone.critic_1_optimizer,
                    clone.critic_2_optimizer,
                ) = self.accelerator.prepare(
                    actor,
                    actor_target,
                    critic_1,
                    critic_target_1,
                    critic_2,
                    critic_target_2,
                    actor_optimizer,
                    critic_1_optimizer,
                    critic_2_optimizer,
                )
            else:
                (
                    clone.actor,
                    clone.actor_target,
                    clone.critic_1,
                    clone.critic_target_1,
                    clone.critic_2,
                    clone.critic_target_2,
                    clone.actor_optimizer,
                    clone.critic_1_optimizer,
                    clone.critic_1_optimizer,
                ) = (
                    actor,
                    actor_target,
                    critic_1,
                    critic_target_1,
                    critic_2,
                    critic_target_2,
                    actor_optimizer,
                    critic_1_optimizer,
                    critic_2_optimizer,
                )
        else:
            clone.actor = actor.to(self.device)
            clone.actor_target = actor_target.to(self.device)
            clone.critic_1 = critic_1.to(self.device)
            clone.critic_target_1 = critic_target_1.to(self.device)
            clone.critic_2 = critic_2.to(self.device)
            clone.critic_target_2 = critic_target_2.to(self.device)
            clone.actor_optimizer = actor_optimizer
            clone.critic_1_optimizer = critic_1_optimizer
            clone.critic_2_optimizer = critic_2_optimizer

        for attribute in self.inspect_attributes().keys():
            if hasattr(self, attribute) and hasattr(clone, attribute):
                attr, clone_attr = getattr(self, attribute), getattr(clone, attribute)
                if isinstance(attr, torch.Tensor) or isinstance(
                    clone_attr, torch.Tensor
                ):
                    if not torch.equal(attr, clone_attr):
                        setattr(
                            clone, attribute, copy.deepcopy(getattr(self, attribute))
                        )
                elif isinstance(attr, np.ndarray) or isinstance(clone_attr, np.ndarray):
                    if not np.array_equal(attr, clone_attr):
                        setattr(
                            clone, attribute, copy.deepcopy(getattr(self, attribute))
                        )
                else:
                    if attr != clone_attr:
                        setattr(
                            clone, attribute, copy.deepcopy(getattr(self, attribute))
                        )
            else:
                setattr(clone, attribute, copy.deepcopy(getattr(self, attribute)))

        if index is not None:
            clone.index = index

        return clone

    def inspect_attributes(self, input_args_only=False):
        # Get all attributes of the current object
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        guarded_attributes = [
            "actor",
            "critic_1",
            "critic_2",
            "actor_target",
            "critic_target_1",
            "critic_target_2",
            "actor_optimizer",
            "critic_1_optimizer",
            "critic_2_optimizer",
        ]

        # Exclude private and built-in attributes
        attributes = [
            a for a in attributes if not (a[0].startswith("__") and a[0].endswith("__"))
        ]

        if input_args_only:
            constructor_params = inspect.signature(self.__init__).parameters.keys()
            attributes = {
                k: v
                for k, v in attributes
                if k not in guarded_attributes and k in constructor_params
            }
        else:
            # Remove the algo specific guarded variables
            attributes = {k: v for k, v in attributes if k not in guarded_attributes}

        return attributes

    def wrap_models(self):
        if self.accelerator is not None:
            (
                self.actor,
                self.actor_target,
                self.critic_1,
                self.critic_target_1,
                self.critic_2,
                self.critic_target_2,
                self.actor_optimizer,
                self.critic_1_optimizer,
                self.critic_2_optimizer,
            ) = self.accelerator.prepare(
                self.actor,
                self.actor_target,
                self.critic_1,
                self.critic_target_1,
                self.critic_2,
                self.critic_target_2,
                self.actor_optimizer,
                self.critic_1_optimizer,
                self.critic_2_optimizer,
            )

    def unwrap_models(self):
        if self.accelerator is not None:
            self.actor = self.accelerator.unwrap_model(self.actor)
            self.actor_target = self.accelerator.unwrap_model(self.actor_target)
            self.critic_1 = self.accelerator.unwrap_model(self.critic_1)
            self.critic_target_1 = self.accelerator.unwrap_model(self.critic_target_1)
            self.critic_2 = self.accelerator.unwrap_model(self.critic_2)
            self.critic_target_2 = self.accelerator.unwrap_model(self.critic_target_2)
            self.actor_optimizer = unwrap_optimizer(
                self.actor_optimizer, self.actor, self.lr_actor
            )
            self.critic_1_optimizer = unwrap_optimizer(
                self.critic_1_optimizer, self.critic_1, self.lr_critic
            )
            self.critic_2_optimizer = unwrap_optimizer(
                self.critic_2_optimizer, self.critic_2, self.lr_critic
            )

    def save_checkpoint(self, path):
        """Saves a checkpoint of agent properties and network weights to path.

        :param path: Location to save checkpoint at
        :type path: string
        """
        attribute_dict = self.inspect_attributes()

        network_info = {
            "actor_init_dict": self.actor.init_dict,
            "actor_state_dict": self.actor.state_dict(),
            "actor_target_init_dict": self.actor_target.init_dict,
            "actor_target_state_dict": self.actor_target.state_dict(),
            "critic_1_init_dict": self.critic_1.init_dict,
            "critic_1_state_dict": self.critic_1.state_dict(),
            "critic_target_1_init_dict": self.critic_target_1.init_dict,
            "critic_target_1_state_dict": self.critic_target_1.state_dict(),
            "critic_2_init_dict": self.critic_2.init_dict,
            "critic_2_state_dict": self.critic_2.state_dict(),
            "critic_target_2_init_dict": self.critic_target_2.init_dict,
            "critic_target_2_state_dict": self.critic_target_2.state_dict(),
            "actor_optimizer_state_dict": self.actor_optimizer.state_dict(),
            "critic_1_optimizer_state_dict": self.critic_1_optimizer.state_dict(),
            "critic_2_optimizer_state_dict": self.critic_2_optimizer.state_dict(),
        }

        attribute_dict.update(network_info)

        torch.save(
            attribute_dict,
            path,
            pickle_module=dill,
        )

    def load_checkpoint(self, path):
        """Loads saved agent properties and network weights from checkpoint.

        :param path: Location to load checkpoint from
        :type path: string
        """
        network_info = [
            "actor_state_dict",
            "actor_target_state_dict",
            "actor_optimizer_state_dict",
            "actor_init_dict",
            "actor_target_init_dict",
            "critic_1_state_dict",
            "critic_target_1_state_dict",
            "critic_1_optimizer_state_dict",
            "critic_1_init_dict",
            "critic_target_1_init_dict",
            "critic_2_state_dict",
            "critic_target_2_state_dict",
            "critic_2_optimizer_state_dict",
            "critic_2_init_dict",
            "critic_target_2_init_dict",
            "net_config",
            "lr_actor",
            "lr_critic",
        ]

        checkpoint = torch.load(path, map_location=self.device, pickle_module=dill)
        self.net_config = checkpoint["net_config"]
        if self.net_config is not None:
            self.arch = checkpoint["net_config"]["arch"]
            if self.arch == "mlp":
                network_class = EvolvableMLP
            elif self.arch == "cnn":
                network_class = EvolvableCNN
        else:
            network_class = MakeEvolvable

        self.actor = network_class(**checkpoint["actor_init_dict"])
        self.actor_target = network_class(**checkpoint["actor_target_init_dict"])
        self.critic_1 = network_class(**checkpoint["critic_1_init_dict"])
        self.critic_target_1 = network_class(**checkpoint["critic_target_1_init_dict"])
        self.critic_2 = network_class(**checkpoint["critic_2_init_dict"])
        self.critic_target_2 = network_class(**checkpoint["critic_target_2_init_dict"])

        self.lr_actor = checkpoint["lr_actor"]
        self.lr_critic = checkpoint["lr_critic"]
        self.actor_optimizer = optim.Adam(self.actor.parameters(), lr=self.lr_actor)
        self.critic_1_optimizer = optim.Adam(
            self.critic_1.parameters(), lr=self.lr_critic
        )
        self.critic_2_optimizer = optim.Adam(
            self.critic_2.parameters(), lr=self.lr_critic
        )
        self.actor.load_state_dict(checkpoint["actor_state_dict"])
        self.actor_target.load_state_dict(checkpoint["actor_target_state_dict"])
        self.critic_1.load_state_dict(checkpoint["critic_1_state_dict"])
        self.critic_target_1.load_state_dict(checkpoint["critic_target_1_state_dict"])
        self.critic_2.load_state_dict(checkpoint["critic_2_state_dict"])
        self.critic_target_2.load_state_dict(checkpoint["critic_target_2_state_dict"])
        self.actor_optimizer.load_state_dict(checkpoint["actor_optimizer_state_dict"])
        self.critic_1_optimizer.load_state_dict(
            checkpoint["critic_1_optimizer_state_dict"]
        )
        self.critic_2_optimizer.load_state_dict(
            checkpoint["critic_2_optimizer_state_dict"]
        )
        for attribute in checkpoint.keys():
            if attribute not in network_info:
                setattr(self, attribute, checkpoint[attribute])

    @classmethod
    def load(cls, path, device="cpu", accelerator=None):
        """Creates agent with properties and network weights loaded from path.

        :param path: Location to load checkpoint from
        :type path: string
        :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
        :type device: str, optional
        :param accelerator: Accelerator for distributed computing, defaults to None
        :type accelerator: accelerate.Accelerator(), optional
        """
        checkpoint = torch.load(path, map_location=device, pickle_module=dill)
        checkpoint["actor_init_dict"]["device"] = device
        checkpoint["actor_target_init_dict"]["device"] = device
        checkpoint["critic_1_init_dict"]["device"] = device
        checkpoint["critic_target_1_init_dict"]["device"] = device
        checkpoint["critic_2_init_dict"]["device"] = device
        checkpoint["critic_target_2_init_dict"]["device"] = device

        actor_init_dict = chkpt_attribute_to_device(
            checkpoint.pop("actor_init_dict"), device
        )
        actor_target_init_dict = chkpt_attribute_to_device(
            checkpoint.pop("actor_target_init_dict"), device
        )
        actor_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("actor_state_dict"), device
        )
        actor_target_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("actor_target_state_dict"), device
        )
        actor_optimizer_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("actor_optimizer_state_dict"), device
        )
        critic_1_init_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_1_init_dict"), device
        )
        critic_target_1_init_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_target_1_init_dict"), device
        )
        critic_1_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_1_state_dict"), device
        )
        critic_target_1_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_target_1_state_dict"), device
        )
        critic_1_optimizer_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_1_optimizer_state_dict"), device
        )
        critic_2_init_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_2_init_dict"), device
        )
        critic_target_2_init_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_target_2_init_dict"), device
        )
        critic_2_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_2_state_dict"), device
        )
        critic_target_2_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_target_2_state_dict"), device
        )
        critic_2_optimizer_state_dict = chkpt_attribute_to_device(
            checkpoint.pop("critic_2_optimizer_state_dict"), device
        )

        checkpoint["device"] = device
        checkpoint["accelerator"] = accelerator
        checkpoint = chkpt_attribute_to_device(checkpoint, device)

        constructor_params = inspect.signature(cls.__init__).parameters.keys()
        class_init_dict = {
            k: v for k, v in checkpoint.items() if k in constructor_params
        }

        if checkpoint["net_config"] is not None:
            agent = cls(**class_init_dict)
            agent.arch = checkpoint["net_config"]["arch"]
            if agent.arch == "mlp":
                agent.actor = EvolvableMLP(**actor_init_dict)
                agent.actor_target = EvolvableMLP(**actor_target_init_dict)
                agent.critic_1 = EvolvableMLP(**critic_1_init_dict)
                agent.critic_target_1 = EvolvableMLP(**critic_target_1_init_dict)
                agent.critic_2 = EvolvableMLP(**critic_2_init_dict)
                agent.critic_target_2 = EvolvableMLP(**critic_target_2_init_dict)
            elif agent.arch == "cnn":
                agent.actor = EvolvableCNN(**actor_init_dict)
                agent.actor_target = EvolvableCNN(**actor_target_init_dict)
                agent.critic_1 = EvolvableCNN(**critic_1_init_dict)
                agent.critic_target_1 = EvolvableCNN(**critic_target_1_init_dict)
                agent.critic_2 = EvolvableCNN(**critic_2_init_dict)
                agent.critic_target_2 = EvolvableCNN(**critic_target_2_init_dict)
        else:
            class_init_dict["actor_network"] = MakeEvolvable(**actor_init_dict)
            class_init_dict["critic_networks"] = [
                MakeEvolvable(**critic_1_init_dict),
                MakeEvolvable(**critic_2_init_dict),
            ]
            agent = cls(**class_init_dict)
            agent.actor_target = MakeEvolvable(**actor_target_init_dict)
            agent.critic_target_1 = MakeEvolvable(**critic_target_1_init_dict)
            agent.critic_target_2 = MakeEvolvable(**critic_target_2_init_dict)

        agent.actor_optimizer = optim.Adam(agent.actor.parameters(), lr=agent.lr_actor)
        agent.actor.load_state_dict(actor_state_dict)
        agent.actor_target.load_state_dict(actor_target_state_dict)
        agent.actor_optimizer.load_state_dict(actor_optimizer_state_dict)

        agent.critic_1_optimizer = optim.Adam(
            agent.critic_1.parameters(), lr=agent.lr_critic
        )
        agent.critic_1.load_state_dict(critic_1_state_dict)
        agent.critic_target_1.load_state_dict(critic_target_1_state_dict)
        agent.critic_1_optimizer.load_state_dict(critic_1_optimizer_state_dict)

        agent.critic_2_optimizer = optim.Adam(
            agent.critic_2.parameters(), lr=agent.lr_critic
        )
        agent.critic_2.load_state_dict(critic_2_state_dict)
        agent.critic_target_2.load_state_dict(critic_target_2_state_dict)
        agent.critic_2_optimizer.load_state_dict(critic_2_optimizer_state_dict)

        if accelerator is not None:
            agent.wrap_models()

        for attribute in agent.inspect_attributes().keys():
            setattr(agent, attribute, checkpoint[attribute])

        return agent



"""An async vector pettingzoo environment"""

import multiprocessing as mp
import operator
import sys
import time
import traceback
from collections import defaultdict
from copy import deepcopy
from enum import Enum
from itertools import accumulate
from typing import TypeVar

import numpy as np
from gymnasium import logger
from gymnasium.error import (
    AlreadyPendingCallError,
    ClosedEnvironmentError,
    NoAsyncCallError,
)
from gymnasium.vector.utils import CloudpickleWrapper, batch_space, clear_mpi_env_vars

from agilerl.vector.pz_vec_env import PettingZooVecEnv

AgentID = TypeVar("AgentID")
ObsType = TypeVar("ObsType")


class AsyncState(Enum):
    """The AsyncVectorEnv possible states given the different actions."""

    DEFAULT = "default"
    WAITING_RESET = "reset"
    WAITING_STEP = "step"
    WAITING_CALL = "call"


class AsyncPettingZooVecEnv(PettingZooVecEnv):
    """Vectorized PettingZoo environment that runs multiple environments in parallel

    :param env_fns: Functions that create the environment
    :type env_fns: list[Callable]
    :param copy: Boolean flag to copy the observation data when it is returned with either .step() or .reset(), recommended, defaults to True
    :type copy: bool, optional
    :param context: Context for multiprocessing
    """

    def __init__(self, env_fns, experience_spec=None, copy=True, context=None):

        # Core class attributes
        self.env_fns = env_fns
        self.num_envs = len(env_fns)
        dummy_env = env_fns[0]()
        self.metadata = (
            dummy_env.metadata
            if hasattr(dummy_env, "metadata")
            else dummy_env.unwrapped.metadata
        )
        self.render_mode = (
            dummy_env.render_mode
            if hasattr(dummy_env, "render_mode")
            else dummy_env.unwrapped.render_mode
        )
        self.possible_agents = dummy_env.possible_agents
        self.copy = copy

        if experience_spec is None:
            self.experience_spec = PettingZooExperienceSpec(self.num_envs)
        else:
            self.experience_spec = experience_spec

        self.experience_spec.detect_space_info(dummy_env)
        del dummy_env
        ctx = mp.get_context(context)

        self.action_space = self._get_action_space
        self.observation_space = self._get_observation_space
        self.single_action_space = self._get_single_action_space
        self.single_observation_space = self._get_single_observation_space

        # Create the shared memory for sharing observations between subprocesses
        self._obs_buffer = SharedMemory(
            num_envs=self.num_envs, exp_spec=self.experience_spec, context=ctx
        )
        self.observations = Observations(
            shared_memory=self._obs_buffer.shared_memory,
            exp_spec=self.experience_spec,
            num_envs=self.num_envs,
        )

        self.parent_pipes, self.processes = [], []
        self.error_queue = ctx.Queue()

        target = _async_worker
        with clear_mpi_env_vars():
            for idx, env_fn in enumerate(self.env_fns):
                parent_pipe, child_pipe = ctx.Pipe()
                process = ctx.Process(
                    target=target,
                    name=f"Worker<{type(self).__name__}>-{idx}",
                    args=(
                        idx,
                        CloudpickleWrapper(env_fn),
                        child_pipe,
                        parent_pipe,
                        self._obs_buffer.shared_memory,
                        self.error_queue,
                        self.experience_spec.observation_shapes,
                        self.experience_spec.observation_widths,
                        self.experience_spec.observation_dtypes,
                        self.experience_spec.agents,
                    ),
                )
                self.parent_pipes.append(parent_pipe)
                self.processes.append(process)
                process.daemon = True
                process.start()
                child_pipe.close()
        self._state = AsyncState.DEFAULT

        super().__init__(
            len(env_fns),
            self.possible_agents,
        )

    def _get_single_action_space(self, agent):
        """Get an agents single action space

        :param agent: Name of agent
        :type agent: str
        """
        return self.experience_spec.single_action_space[agent]

    def _get_action_space(self, agent):
        """Get an agents action space

        :param agent: Name of agent
        :type agent: str
        """
        return self.experience_spec.action_space[agent]

    def _get_single_observation_space(self, agent):
        """Get an agents single observation space

        :param agent: Name of agent
        :type agent: str
        """
        return self.experience_spec.single_observation_space[agent]

    def _get_observation_space(self, agent):
        """Get an agents observation space

        :param agent: Name of agent
        :type agent: str
        """
        return self.experience_spec.observation_space[agent]

    def reset(
        self,
        *,
        seed=None,
        options=None,
    ):
        """
        Reset all the environments and return two dictionaries of batched observations and infos.

        :param seed: Random seed, defaults to None
        :type seed: None | int, optional
        :param options: Options dictionary
        :type options: dict[str, Any]
        """
        self.reset_async(seed=seed, options=options)
        return self.reset_wait()

    def reset_async(
        self,
        seed=None,
        options=None,
    ):
        """Send calls to the :obj:`reset` methods of the sub-environments.

        To get the results of these calls, you may invoke :meth:`reset_wait`.

        :param seed: Random seed, defaults to None
        :type seed: None | int, optional
        :param options: Options dictionary
        :type options: dict[str, Any]
        """
        self._assert_is_running()
        if seed is None:
            seed = [None for _ in range(self.num_envs)]
        elif isinstance(seed, int):
            seed = [seed + i for i in range(self.num_envs)]
        assert (
            len(seed) == self.num_envs
        ), f"If seeds are passed as a list the length must match num_envs={self.num_envs} but got length={len(seed)}."

        if self._state != AsyncState.DEFAULT:
            raise AlreadyPendingCallError(
                f"Calling `reset_async` while waiting for a pending call to `{self._state.value}` to complete",
                str(self._state.value),
            )
        for pipe, env_seed in zip(self.parent_pipes, seed):
            env_kwargs = {"seed": env_seed, "options": options}
            pipe.send(("reset", env_kwargs))
        self._state = AsyncState.WAITING_RESET

    def reset_wait(self, timeout=None):
        """Waits for the calls triggered by :meth:`reset_async` to finish and returns the results.

        :param timeout: Number of seconds before the call to ``reset_wait`` times out. If `None`, the call to ``reset_wait`` never times out, defaults to 0
        :type timeout: int | float | None, optional
        """
        self._assert_is_running()
        if self._state != AsyncState.WAITING_RESET:
            raise NoAsyncCallError(
                "Calling `reset_wait` without any prior " "call to `reset_async`.",
                AsyncState.WAITING_RESET.value,
            )

        if not self._poll_pipe_envs(timeout):
            self._state = AsyncState.DEFAULT
            raise mp.TimeoutError(
                f"The call to `reset_wait` has timed out after {timeout} second(s)."
            )

        info_data, successes = zip(*[pipe.recv() for pipe in self.parent_pipes])
        self._raise_if_errors(successes)

        infos = {}
        # Convert info data to list before passing into key_search function
        for i, info in enumerate(info_data):
            infos = self._add_info(infos, info, i)

        self._state = AsyncState.DEFAULT
        return (
            (
                {
                    agent: deepcopy(self.observations[agent])
                    for agent in self.observations.keys()
                }
                if self.copy
                else self.observations
            ),
            infos,
        )

    def step_async(self, actions):
        """
        Tell all the environments to start taking a step
        with the given actions.
        Call step_wait() to get the results of the step.
        You should not call this if a step_async run is
        already pending.

        :param actions: List of lists of length num_envs, each sub list contains actions for each agent in a given environment
        :type actions: list[list[int | float | np.ndarray]]
        """
        self._assert_is_running()
        if self._state != AsyncState.DEFAULT:
            raise AlreadyPendingCallError(
                f"Calling `step_async` while waiting for a pending call to `{self._state.value}` to complete.",
                str(self._state.value),
            )
        for pipe, action in zip(self.parent_pipes, actions):
            pipe.send(("step", action))
        self._state = AsyncState.WAITING_STEP

    def step_wait(self, timeout=None):
        """
        Wait for the calls to :obj:`step` in each sub-environment to finish.

        :param timeout: Number of seconds before the call to ``step_wait`` times out. If `None`, the call to ``step_wait`` never times out, defaults to 0
        :type timeout: int | float | None, optional
        """

        self._assert_is_running()
        if self._state != AsyncState.WAITING_STEP:
            raise NoAsyncCallError(
                "Calling `step_wait` without any prior call " "to `step_async`.",
                AsyncState.WAITING_STEP.value,
            )

        if not self._poll_pipe_envs(timeout):
            self._state = AsyncState.DEFAULT
            raise mp.TimeoutError(
                f"The call to `step_wait` has timed out after {timeout} second(s)."
            )

        rewards, terminations, truncations, infos = (
            defaultdict(list) for _ in range(4)
        )
        successes = []
        infos = {}
        for env_idx, pipe in enumerate(self.parent_pipes):
            env_step_return, success = pipe.recv()
            successes.append(success)
            if success:
                for agent in self.agents:
                    rewards[agent].append(env_step_return[0][agent])
                    terminations[agent].append(env_step_return[1][agent])
                    truncations[agent].append(env_step_return[2][agent])
                infos = self._add_info(infos, env_step_return[3], env_idx)

        self._raise_if_errors(successes)
        self._state = AsyncState.DEFAULT
        return (
            (
                {
                    agent: deepcopy(self.observations[agent])
                    for agent in self.observations.keys()
                }
                if self.copy
                else self.observations
            ),
            {agent: np.array(rew) for agent, rew in rewards.items()},
            {agent: np.array(term) for agent, term in terminations.items()},
            {agent: np.array(trunc) for agent, trunc in truncations.items()},
            infos,
        )

    def render(self):
        """
        Returns the rendered frames from the parallel environments.

        """

        return self.call("render")

    def call(self, name, *args, **kwargs):
        """
        Call a method from each parallel environment with args and kwargs.

        :param name: Name of the method or property to call
        :type name: str
        :param *args: Position arguments to apply to the method call.
        :type *args: Any
        :param **kwargs: Keyword arguments to apply to the method call.
        :type **kwargs: Any
        """
        self.call_async(name, *args, **kwargs)
        return self.call_wait()

    def call_async(self, name, *args, **kwargs):
        """
        Calls the method with name asynchronously and apply args and kwargs to the method.

        :param name: Name of the method or property to call
        :type name: str
        :param *args: Position arguments to apply to the method call.
        :type *args: Any
        :param **kwargs: Keyword arguments to apply to the method call.
        :type **kwargs: Any
        """
        self._assert_is_running()
        if self._state != AsyncState.DEFAULT:
            raise AlreadyPendingCallError(
                f"Calling `call_async` while waiting for a pending call to `{self._state.value}` to complete.",
                str(self._state.value),
            )

        for pipe in self.parent_pipes:
            pipe.send(("_call", (name, args, kwargs)))
        self._state = AsyncState.WAITING_CALL

    def call_wait(self, timeout=None):
        """Calls all parent pipes and waits for the results.

        :param timeout: Number of seconds before the call to :meth:`call_wait` times out. If ``None`` (default), the call to :meth:`call_wait` never times out, defaults to 0
        :type timeout: int | float | None, optional
        """
        self._assert_is_running()
        if self._state != AsyncState.WAITING_CALL:
            raise NoAsyncCallError(
                "Calling `call_wait` without any prior call to `call_async`.",
                AsyncState.WAITING_CALL.value,
            )

        if not self._poll_pipe_envs(timeout):
            self._state = AsyncState.DEFAULT
            raise mp.TimeoutError(
                f"The call to `call_wait` has timed out after {timeout} second(s)."
            )

        results, successes = zip(*[pipe.recv() for pipe in self.parent_pipes])
        self._raise_if_errors(successes)
        self._state = AsyncState.DEFAULT
        return results

    def get_attr(self, name):
        """
        Get a property from each parallel environment.

        :param name: Name of property to get from each individual environment
        :type name: str
        """
        return self.call(name)

    def set_attr(self, name, values):
        """Sets an attribute of the sub-environments.

        :param name: Name of the property to be set in each individual environment.
        :type name: str
        :param values: Values of the property to be set to. If ``values`` is a list or
            tuple, then it corresponds to the values for each individual
            environment, otherwise a single value is set for all environments.
        :type values: list[Any] | tuple[Any] | object
        """
        self._assert_is_running()
        if not isinstance(values, (list, tuple)):
            values = [values for _ in range(self.num_envs)]
        if len(values) != self.num_envs:
            raise ValueError(
                "Values must be a list or tuple with length equal to the number of environments. "
                f"Got `{len(values)}` values for {self.num_envs} environments."
            )

        if self._state != AsyncState.DEFAULT:
            raise AlreadyPendingCallError(
                f"Calling `set_attr` while waiting for a pending call to `{self._state.value}` to complete.",
                str(self._state.value),
            )

        for pipe, value in zip(self.parent_pipes, values):
            pipe.send(("_setattr", (name, value)))
        _, successes = zip(*[pipe.recv() for pipe in self.parent_pipes])
        self._raise_if_errors(successes)

    def close_extras(self, timeout=None, terminate=False):
        """
        Close the environments & clean up the extra resources (processes and pipes).

        :param timeout: Number of seconds before the call to :meth:`close` times out. If ``None``,
                the call to :meth:`close` never times out. If the call to :meth:`close`
                times out, then all processes are terminated, defaults to 0
        :type timeout: int | float | None, optional
        :param terminate: If ``True``, then the :meth:`close` operation is forced and all processes are terminated, defaults to False
        :type terminate: bool, optional
        """
        timeout = 0 if terminate else timeout

        try:
            if self._state != AsyncState.DEFAULT:
                logger.warn(
                    f"Calling `close` while waiting for a pending call to `{self._state.value}` to complete."
                )
                function = getattr(self, f"{self._state.value}_wait")
                function(timeout)
        except mp.TimeoutError:
            terminate = True

        if terminate:
            for process in self.processes:
                if process.is_alive():
                    process.terminate()
        else:
            for pipe in self.parent_pipes:
                if (pipe is not None) and (not pipe.closed):
                    pipe.send(("close", None))

            for pipe in self.parent_pipes:
                if (pipe is not None) and (not pipe.closed):
                    pipe.recv()

        for pipe in self.parent_pipes:
            if pipe is not None:
                pipe.close()
        for process in self.processes:
            process.join()

    def _poll_pipe_envs(self, timeout: int | None = None):
        self._assert_is_running()

        if timeout is None:
            return True

        end_time = time.perf_counter() + timeout
        for pipe in self.parent_pipes:
            delta = max(end_time - time.perf_counter(), 0)

            if pipe is None:
                return False
            if pipe.closed or (not pipe.poll(delta)):
                return False
        return True

    def _raise_if_errors(self, successes: list[bool] | tuple[bool]):
        if all(successes):
            return

        num_errors = self.num_envs - sum(successes)
        assert num_errors > 0
        for i in range(num_errors):
            index, exctype, value, trace = self.error_queue.get()

            logger.error(
                f"Received the following error from Worker-{index} - Shutting it down"
            )
            logger.error(f"{trace}")

            self.parent_pipes[index].close()
            self.parent_pipes[index] = None

            if i == num_errors - 1:
                logger.error("Raising the last exception back to the main process.")
                self._state = AsyncState.DEFAULT
                raise exctype(value)

    def _assert_is_running(self):
        if self.closed:
            raise ClosedEnvironmentError(
                f"Trying to operate on `{type(self).__name__}`, after a call to `close()`."
            )

    def _add_info(self, vector_infos, env_info, env_num):
        """
        Compile a vectorized information dictionary
        """

        for key, value in env_info.items():
            # If value is a dictionary, then we apply the `_add_info` recursively.
            if isinstance(value, dict):
                array = self._add_info(vector_infos.get(key, {}), value, env_num)
            # Otherwise, we are a base case to group the data
            else:
                # If the key doesn't exist in the vector infos, then we can create an array of that batch type
                if key not in vector_infos:
                    if type(value) in [int, float, bool] or issubclass(
                        type(value), np.number
                    ):
                        array = np.zeros(self.num_envs, dtype=type(value))
                    elif isinstance(value, np.ndarray):
                        # We assume that all instances of the np.array info are of the same shape
                        array = np.zeros(
                            (self.num_envs, *value.shape), dtype=value.dtype
                        )
                    elif value is None:
                        array = np.full(
                            self.num_envs, fill_value=np.nan, dtype=np.float32
                        )
                    else:
                        # For unknown objects, we use a Numpy object array
                        array = np.full(self.num_envs, fill_value=None, dtype=object)

                # Otherwise, just use the array that already exists
                else:
                    array = vector_infos[key]

                # Assign the data in the `env_num` position
                #   We only want to run this for the base-case data (not recursive data forcing the ugly function structure)
                array[env_num] = value

            # Get the array mask and if it doesn't already exist then create a zero bool array
            array_mask = vector_infos.get(
                f"_{key}", np.zeros(self.num_envs, dtype=np.bool_)
            )
            array_mask[env_num] = True

            # Update the vector info with the updated data and mask information
            vector_infos[key], vector_infos[f"_{key}"] = array, array_mask

        return vector_infos

    def __del__(self):
        """On deleting the object, checks that the vector environment is closed."""
        if not getattr(self, "closed", True) and hasattr(self, "_state"):
            self.close(terminate=True)
        if hasattr(self, "_obs_buffer"):
            del self._obs_buffer
        if hasattr(self, "observations"):
            del self.observations
        if hasattr(self, "experience_spec"):
            del self.experience_spec


class PettingZooExperienceSpec:
    """Class for formatting experiences when being returned by a vectorized environment

    :param num_envs: Number of environments to vectorize
    :type num_envs: int
    """

    def __init__(self, num_envs):
        self.num_envs = num_envs

    def detect_space_info(self, dummy_env):

        self.metadata = (
            dummy_env.metadata
            if hasattr(dummy_env, "metadata")
            else dummy_env.unwrapped.metadata
        )
        self.render_mode = (
            dummy_env.render_mode
            if hasattr(dummy_env, "render_mode")
            else dummy_env.unwrapped.render_mode
        )
        self.possible_agents = dummy_env.possible_agents

        try:
            # Collect action space data
            self.single_action_space = {
                agent: dummy_env.action_space(agent)
                for agent in dummy_env.possible_agents
            }
            self.action_space = {
                agent: batch_space(self.single_action_space[agent], self.num_envs)
                for agent in dummy_env.possible_agents
            }

            # Collect observation space data
            self.single_observation_space = {
                agent: dummy_env.observation_space(agent)
                for agent in dummy_env.possible_agents
            }

            self.observation_dtypes = {
                agent: dummy_env.observation_space(agent).dtype
                for agent in dummy_env.possible_agents
            }

            self.observation_space = {
                agent: batch_space(self.single_observation_space[agent], self.num_envs)
                for agent in dummy_env.possible_agents
            }

            self.observation_shapes = {
                agent: space.shape if space.shape != () else (1,)
                for agent, space in self.single_observation_space.items()
            }

            # Width of each agents flattened observations
            self.observation_widths = {
                agent: int(np.prod(obs.shape))
                for agent, obs in self.single_observation_space.items()
            }
            # Cumulative widths of the flattened observations
            self.observation_boundaries = [0] + list(
                accumulate(self.observation_widths.values(), operator.add)
            )
        except Exception as e:
            raise ValueError(
                "Unable to calculate observation dimensions from current observation space type."
            ) from e

        # Total width of all agents observations, flattened and concatenated
        self.total_observation_width = int(
            np.sum(list(self.observation_widths.values()))
        )
        dummy_env.reset()

        # This is the gospel order of agents
        self.agents = dummy_env.possible_agents[:]
        self.agent_index_map = {agent: i for i, agent in enumerate(self.agents)}

        dummy_env.close()
        del dummy_env


class Observations:
    """
    Class for storing observations with a dictionary interface

    :param shared_memory: A RawArray that all envs write observations to.
    :type shared_memory: multiprocessing.RawArray
    :param exp_spec: Experience specification
    :type exp_spec: agilerl.vector.pz_async_vec_env.PettingZooExperienceSpec
    :param num_envs: Number of environments
    :type num_envs: int

    """

    def __init__(self, shared_memory, exp_spec, num_envs):
        self.exp_spec = exp_spec
        self.num_envs = num_envs
        self.shared_memory = shared_memory
        self.obs_view = []
        for shm, agent in zip(shared_memory, exp_spec.agents):
            self.obs_view.append(
                np.frombuffer(
                    shm.get_obj(), dtype=exp_spec.single_observation_space[agent].dtype
                )
            )

    def __getitem__(self, key):
        """
        Get agent observation given a key (agent_id)
        """
        agent_idx = self.exp_spec.agent_index_map[key]
        return self.obs_view[agent_idx].reshape(
            (self.num_envs, *self.exp_spec.observation_shapes[key])
        )

    def __str__(self):
        """"""
        my_dic = {agent: obs for agent, obs in zip(self.exp_spec.agents, self.obs_view)}
        return f"{my_dic}"

    def __repr__(self):
        return self.__str__()

    def __contains__(self, key):  # key is agent_id
        return key in self.exp_spec.agent_index_map.keys()

    def __len__(self):
        return len(self.exp_spec.agents)

    def __iterate_kv(self):
        for key in self.exp_spec.agents:
            yield (key, self.__getitem__(key))

    def __iter__(self):
        return self.__iterate_kv()

    def keys(self):
        for k, _ in self.__iterate_kv():
            yield k

    def values(self):
        for _, v in self.__iterate_kv():
            yield v

    def items(self):
        return self.__iterate_kv()

    def get(self, key):
        try:
            return self.__getitem__(key)
        except KeyError:
            return None


class SharedMemory:
    """Class to hold the shared memory object that each of the subprocesses will write their observation to.

    :param num_envs: Number of environments to vectorize
    :type num_envs: int
    :param exp_spec: Experience specification
    :type exp_spec: PettingZooExperienceSpec
    :param context: Multiprocessing context
    :type context: BaseContext

    """

    def __init__(self, num_envs, exp_spec, context):
        self.shared_memory = []
        for agent in exp_spec.agents:
            shared_memory = context.Array(
                exp_spec.single_observation_space[agent].dtype.char,
                exp_spec.observation_widths[agent] * num_envs,
            )
            self.shared_memory.append(shared_memory)


def get_placeholder_value(agent, transition_name, observation_shapes=None):
    """When an agent is killed, used to obtain a placeholder value to return for associated experience.

    :param agent: Agent ID
    :type agent: str
    :param transition_name: Name of the transition
    :type transition_name: str
    :param observations: Observations numpy array backed by RawArray, defaults to None
    :type observations: agilerl.vector.pz_async_vec_env.Observations, optional
    """
    match transition_name:
        case "reward":
            return 0
        case "truncated":
            return False
        case "terminated":
            return True
        case "info":
            return {}
        case "observation":
            return -np.ones_like(observation_shapes[agent])


def process_transition(transitions, observation_shapes, transition_names, agents):
    """Process transition, adds in placeholder values for killed sub-agents

    :param transitions: Tuple of environment transition
    :type transitions: Tuple[Any]
    :param observation_shapes: Observation shapes
    :type observation_shapes: Dict[str, Tuple[int]]
    :param transition_names: Names associated to transitions
    :type transition_names: List[str]
    :param agents: List of sub-agent names
    :type agents: List[str]
    """
    transition_list = list(transitions)
    for transition, name in zip(transition_list, transition_names):
        transition = {
            agent: (
                transition[agent]
                if agent in transition.keys()
                else get_placeholder_value(agent, name, observation_shapes)
            )
            for agent in agents
        }
    return transition_list


def set_env_obs(index, observation, shared_memory, widths, dtypes):
    """Set the observation for a given environment

    :param index: Environment index
    :type index: int
    :param observation: Observation from env.step or env.reset
    :type observation: Dict[str, np.ndarray]
    :param shared_memory: Shared memory
    :type shared_memory: mp.Array
    :param widths: Flattened observation widths
    :type widths: Dict[str, int]
    :param dtypes: Observation dtypes
    :type dtypes: Dict[str, np.dtype]
    """
    for agent_idx, (agent, obs) in enumerate(observation.items()):
        dest = np.frombuffer(shared_memory[agent_idx].get_obj(), dtype=dtypes[agent])
        np.copyto(
            dest[index * widths[agent] : (index + 1) * widths[agent]],
            np.asarray(obs, dtype=dtypes[agent]).flatten(),
        )


def _async_worker(
    index,
    env_fn,
    pipe,
    parent_pipe,
    shared_memory,
    error_queue,
    observation_shapes,
    observation_widths,
    observation_dtypes,
    agents,
):
    """
    :param index: Subprocess index
    :type index: int
    :param env_fn: Function to call environment
    :type env_fn: callable
    :param pipe: Child pipe object for sending data to the main process
    :type pipe: Connection
    :param parent_pipe: Parent pipe object
    :type parent_pipe: Connection
    :param shared_memory: List of shared memories.
    :type shared_memory: List[multiprocessing.Array]
    :param error_queue: Queue object for collecting subprocess errors to communicate back to the main process
    :type error_queue: mp.Queue
    :param observation_shapes: Shapes of observations
    :type observation_shapes: Dict[str, Tuple[int]]
    :param observation_widths: Flattened observation widths
    :type observation_widths: Dict[str, int]
    :param observation_dtypes: Observation dtypes
    :type observation_dtypes: Dict[str, np.dtype]
    :param agents: Sub-agent names
    :type agents: str

    """
    env = env_fn()
    autoreset = False
    parent_pipe.close()

    try:
        while True:
            command, data = pipe.recv()
            if command == "reset":
                observation, info = process_transition(
                    env.reset(**data),
                    observation_shapes,
                    ["observation", "info"],
                    agents,
                )
                set_env_obs(
                    index,
                    observation,
                    shared_memory,
                    observation_widths,
                    observation_dtypes,
                )
                autoreset = False
                pipe.send(((info), True))
            elif command == "step":
                if autoreset:
                    observation, info = process_transition(
                        env.reset(), observation_shapes, ["observation", "info"], agents
                    )
                    reward = {agent: 0 for agent in agents}
                    terminated = {agent: False for agent in agents}
                    truncated = {agent: False for agent in agents}
                else:
                    data = {
                        possible_agent: (
                            np.array(data[idx]).squeeze()
                            if not isinstance(data[idx], int)
                            else data[idx]
                        )
                        for idx, possible_agent in enumerate(agents)
                    }
                    transition = env.step(data)
                    observation, reward, terminated, truncated, info = transition
                    observation, reward, terminated, truncated, info = (
                        process_transition(
                            transition,
                            observation_shapes,
                            [
                                "observation",
                                "reward",
                                "terminated",
                                "truncated",
                                "info",
                            ],
                            agents,
                        )
                    )
                autoreset = all(
                    [
                        term | trunc
                        for term, trunc in zip(terminated.values(), truncated.values())
                    ]
                )
                set_env_obs(
                    index,
                    observation,
                    shared_memory,
                    observation_widths,
                    observation_dtypes,
                )
                pipe.send(((reward, terminated, truncated, info), True))

            elif command == "close":
                pipe.send((None, True))
                break
            elif command == "_call":
                name, args, kwargs = data
                if name in ["reset", "step", "close", "_setattr", "_check_spaces"]:
                    raise ValueError(
                        f"Trying to call function `{name}` with `call`, use `{name}` directly instead."
                    )
                attr = getattr(env, name)
                if callable(attr):
                    pipe.send((attr(*args, **kwargs), True))
                else:
                    pipe.send((attr, True))
            elif command == "_setattr":
                name, value = data
                setattr(env, name, value)
                pipe.send((None, True))
            else:
                raise RuntimeError(
                    f"Received unknown command `{command}`. Must be one of [`reset`, `step`, `close`, `_call`, `_setattr`, `_check_spaces`]."
                )

    except (KeyboardInterrupt, Exception):
        error_type, error_message, _ = sys.exc_info()
        trace = traceback.format_exc()
        error_queue.put((index, error_type, error_message, trace))
        pipe.send((None, False))

    finally:
        env.close()



import warnings

from torch.utils.data import DataLoader

from agilerl.components.multi_agent_replay_buffer import MultiAgentReplayBuffer
from agilerl.components.replay_buffer import (
    MultiStepReplayBuffer,
    PrioritizedReplayBuffer,
    ReplayBuffer,
)
from agilerl.components.replay_data import ReplayDataset


class Sampler:
    """Sampler class to handle both standard and distributed training."""

    def __init__(
        self,
        distributed=False,
        per=False,
        n_step=False,
        memory=None,
        dataset=None,
        dataloader=None,
    ):
        assert (memory is not None) or (
            (dataset is not None) and (dataloader is not None)
        ), "Sampler needs to be initialized with either 'memory' or ('dataset' AND 'dataloader')."

        self.distributed = distributed
        self.per = per
        self.n_step = n_step
        self.memory = memory
        self.dataset = dataset
        self.dataloader = dataloader

        if self.distributed:
            if not isinstance(self.dataset, ReplayDataset):
                warnings.warn("Dataset is not an agilerl ReplayDataset.")
            if not isinstance(self.dataloader, DataLoader):
                warnings.warn("Dataset is not a torch DataLoader object.")
            self.sample = self.sample_distributed

        elif self.per:
            if not isinstance(self.memory, PrioritizedReplayBuffer):
                warnings.warn("Memory is not an agilerl PrioritizedReplayBuffer.")
            self.sample = self.sample_per

        elif self.n_step:
            if not isinstance(self.memory, MultiStepReplayBuffer):
                warnings.warn("Memory is not an agilerl MultiStepReplayBuffer.")
            self.sample = self.sample_n_step

        else:
            if not isinstance(self.memory, (ReplayBuffer, MultiAgentReplayBuffer)):
                warnings.warn(
                    "Memory is not an agilerl ReplayBuffer or MultiAgentReplayBuffer."
                )
            self.sample = self.sample_standard

    def sample_standard(self, batch_size, return_idx=False):
        return self.memory.sample(batch_size, return_idx)

    def sample_distributed(self, batch_size, return_idx=None):
        self.dataset.batch_size = batch_size
        return next(iter(self.dataloader))

    def sample_per(self, batch_size, beta):
        return self.memory.sample(batch_size, beta)

    def sample_n_step(self, idxs):
        return self.memory.sample_from_indices(idxs)
